import unittest

import numpy as np

from metermetrics import helper


class HelperTest(unittest.TestCase):
    def test_mae(self):
        pred = np.array([1, 2, 3, 4, 5, 6])
        true = np.array([0, 2, 3, 4, 5, 6])

    def test_err(self):
        pred = 810.1
        true = 810.0
        assert helper.err(pred, true) == 0.1

        pred = np.array([0, 0, 1, 1, 1, 1])
        true = np.array([[0], [0], [1], [1], [1], [1]])
        assert helper.err(pred, true) == 0

        pred = np.array([0, 0, 1, 1, 1, 1])
        true = np.array([0, 0, 1, 1, 1, 1])
        assert helper.err(pred, true) == 0

        pred = np.array([0, 0, 1, 1, 0, 1])
        true = np.array([0, 0, 1, 1, 1, 1])
        assert helper.err(pred, true) == -1, helper.err(pred, true)

        pred = np.array([0, 0, 1, 1, 2, 1])
        true = np.array([0, 0, 1, 1, 1, 1])
        assert helper.err(pred, true) == 1, helper.err(pred, true)

        pred = 1
        true = 2
        assert helper.err(pred, true) == -0.1, helper.err(pred, true)

        # 003370, 004370
        pred = np.array([0.0, 0.0, 3.212705, 2.7218645, 7.1418715, 0.0])
        true = np.array([0.0, 0.0, 4.0, 3.0, 7.0, 0.0])
        assert helper.err(pred, true) == -100.0, helper.err(pred, true)

    def test_equal(self):
        pred, true = 1.0, 2.0

        assert not helper.equal(pred, true, strict=1)
        assert not helper.equal(pred, true, strict=0)

        pred, true = 1.0, 0.1
        assert not helper.equal(pred, true, strict=1)
        assert helper.equal(pred, true, strict=0)

        pred, true = 1.0, 1.1
        assert helper.equal(pred, true, strict=1)
        assert helper.equal(pred, true, strict=0)

        pred, true = 1.1, 1.1
        assert helper.equal(pred, true, strict=1)
        assert helper.equal(pred, true, strict=0)

        pred = np.array([1, 2, 3, 4, 5, 6])
        true = np.array([0, 2, 3, 4, 5, 6])
        assert not helper.equal(pred, true, strict=1)
        assert not helper.equal(pred, true, strict=0)

        pred = np.array([1, 2, 4, 4, 5, 5])
        true = np.array([1, 2, 4, 4, 5, 6])
        assert helper.equal(pred, true, strict=1)
        assert helper.equal(pred, true, strict=0)

        pred = np.array([1, 2, 4, 4, 6, 5])
        true = np.array([1, 2, 4, 4, 5, 6])
        assert not helper.equal(pred, true, strict=1)
        assert helper.equal(pred, true, strict=0)

        pred = np.array([0.0, 0.0, 3.212705, 2.7218645, 7.1418715, 0.0])
        true = np.array([0.0, 0.0, 4.0, 3.0, 7.0, 0.0])
        assert not helper.equal(pred, true, strict=0)

    def test_data_uniform(self):
        p = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        p = helper.data_uniform(p)
        assert p == 0

        p = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0], dtype="float32")
        p = helper.data_uniform(p)
        assert p == 0

        pred = np.array([0.0, 0.0, 3.212705, 2.7218645, 7.1418715, 0.0])
        p = helper.data_uniform(pred)
        assert p == 3370, p

    def test_transfer_to_normal(self):
        a = np.array(
            [
                [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            ]
        )
        a = np.concatenate([a.copy().reshape(1, 6, -1) for i in range(11)])
        self.assertEqual(a.shape, (11, 6, 10))

        res = helper.transfer_to_normal(a)
        self.assertEqual(res.shape, (11, 6))



# coding: utf-8
"""
评价标准

"""

import numpy as np


def transfer_to_normal(res: np.array):
    """
    one-hot 数据转换
    :param res:
    :return:
    """
    assert res.shape[0] > 10, "size of dataset must bigger than 10"
    final = np.ones((res.shape[0], 6))
    for i in range(res.shape[0]):
        final[i] = np.argmax(res[i], axis=1)

    return final


def equal(pred, true, strict):
    """
    正常数据，最后一位误差不算作误差.
    <= 1, equal
    <= 10 and not strict, equal

    :param pred:
    :param true:
    :param strict: 最后一位误差如何判断
    :return:
    """

    # 008511 without point
    if abs(err(pred, true)) <= 0.1:
        return True
    if abs(err(pred, true)) < 1 and not strict:
        return True

    return False


def data_uniform(pred):
    """
    trans all kinds of data to int
    :param pred:
    :return:
    """
    if isinstance(pred, str):
        pred = int(pred)
    elif isinstance(pred, np.ndarray):
        if len(pred.shape) > 1:
            pred = pred.reshape(-1)
        if (
            pred.dtype == "float32"
            or pred.dtype == np.ndarray
            or pred.dtype == "float64"
        ):
            pred = np.around(pred)
            pred = pred.astype("int")

        pred = int("".join(pred.astype("str")))
    elif isinstance(pred, float):
        pred = int(pred * 10)
    return pred


def err(pred, true):
    """
    data must be:

    - 008101
    - 0, 0, 8, 1, 0, 0
    - 810.1

    :param pred:
    :param true:
    :return:
    """
    assert type(pred) == type(true)

    pred = data_uniform(pred)
    true = data_uniform(true)

    return (pred - true) / 10


def is_equal_combine(pred, true, strict=False):
    e1 = equal(pred[:-1], true[:-1], strict)
    e2 = equal(pred[-1], true[-1], strict)

    return e1 and e2


def mae_combine(pred, true):
    def mae(pred, true):
        """
        计算误差 MAE

        :param pred:
        :param true:
        :return:
        """

        assert pred.shape[0] == 6

        cnt = 5
        delta = 0

        # pred.shape = (6, )
        for i, j in zip(pred, true):
            assert cnt >= 0
            delta += int((i - j) * (10**cnt))
            cnt -= 1

        return abs(delta)

    return mae(pred[:-1], true[:-1])


def useFileName(cdf_filename):
    def get_filename(epoch):
        filename = cdf_filename % epoch
        return filename

    return get_filename



from .helper import equal, err


class Counter:
    def __init__(self, epoch):
        self.epoch = epoch
        self.pred_list = []
        self.true_list = []
        self.err_list = []
        self.cnt = 0

    def add_val(self, pred, true):
        self.pred_list.append(pred)
        self.true_list.append(true)
        self.err_list.append(err(pred, true))
        self.cnt += 1

    def get_last_err(self, is_abs):
        if is_abs:
            return abs(self.err_list[-1])
        else:
            return self.err_list[-1]

    def is_cur_equal(self, strict):
        pred = self.pred_list[-1]
        true = self.true_list[-1]
        return equal(pred, true, strict)

    def get_mae(self):
        err_list = [abs(val) for val in self.err_list]
        sums = sum(err_list)
        if not err_list:
            return 0
        assert self.cnt == len(self.pred_list)
        return sums / self.cnt

    def get_mse(self):
        sums = 0

        for err_val in self.err_list:
            sums += err_val**2

        if sums == 0:
            return 0
        return sums / self.cnt

    def get_rmse(self):
        return self.get_mse() ** 0.5

    def get_max_error(self):
        err_list = [val for val in self.err_list]
        return max(err_list)

    def get_max_abs_err(self):
        if not self.err_list:
            return -1
        err_list = [abs(val) for val in self.err_list]
        return max(err_list)

    def get_strict_acc(self):
        return cal_acc(self.err_list, strict=1)

    def get_acc(self):
        return cal_acc(self.err_list, strict=0)

    def metrics_to_dict(self):
        row = dict(
            epoch=self.epoch,
            normal_acc=self.get_acc(),
            strict_acc=self.get_strict_acc(),
            MSE=self.get_mse(),
            MAE=self.get_mae(),
            RMSE=self.get_rmse(),
            MAX=self.get_max_abs_err(),
        )
        return row


def cal_acc(err_list: list, strict: bool):
    cnt = 0
    for val in err_list:
        if abs(val) <= 0.1:
            cnt += 1
        elif abs(val) < 1.0 and not strict:
            cnt += 1

    if cnt == 0:
        return 0
    return cnt / len(err_list)



import unittest

from metermetrics.counter import Counter, cal_acc


class CounterTest(unittest.TestCase):
    def setUp(self) -> None:
        self.c = Counter(0)

    def test_counter(self):
        pred_list = [1, 1, 1, 1, 1, 1, 1]
        true_list = [1, 1, 1, 1, 1, 1, 1]
        c = Counter(0)

        for pred, true in zip(pred_list, true_list):
            c.add_val(pred, true)

        assert c.get_mae() == 0
        assert c.get_mse() == 0
        assert c.get_rmse() == 0
        assert c.get_max_error() == 0

    def test_cur_equal(self):
        c = Counter(0)
        c.add_val(2.0, 1.0)
        assert not c.is_cur_equal(strict=0)

        c.add_val(1.0, 1.0)
        assert c.is_cur_equal(1)

        c.add_val(1.1, 1.0)
        assert c.is_cur_equal(1)

        c.add_val(1.2, 1.0)
        assert not c.is_cur_equal(1)

    def test_max_error_and_max_abs_error(self):
        pred_list = [4, 1, 2, 2, 2, 1, 1]
        true_list = [1, 4, 1, 8, 3, 1, 1]

        c = Counter(0)
        for pred, true in zip(pred_list, true_list):
            c.add_val(pred, true)

        assert c.get_max_error() == 0.3, c.get_max_error()
        assert c.get_max_abs_err() == 0.6, c.get_max_error()

        pred_list = [4.0, 1.0, 2.0, 2.0, 2.0, 1.0, 1.0]
        true_list = [1.0, 4.0, 1.0, 8.0, 3.0, 1.0, 1.0]

        c = Counter(0)
        for pred, true in zip(pred_list, true_list):
            c.add_val(pred, true)

        assert c.get_max_error() == 3.0, c.get_max_error()
        assert c.get_max_abs_err() == 6.0, c.get_max_error()

    def test_last_err(self):
        self.c.add_val(1.0, 2.0)
        assert self.c.get_last_err(is_abs=1) == 1.0
        assert self.c.get_last_err(is_abs=0) == -1.0

        self.c.add_val(1.1, 2.0)
        assert self.c.get_last_err(is_abs=1) == 0.9
        assert self.c.get_last_err(is_abs=0) == -0.9

        self.c = Counter(0)
        self.c.add_val(1, 2)
        assert self.c.get_last_err(is_abs=1) == 0.1, self.c.get_last_err(is_abs=1)
        assert self.c.get_last_err(is_abs=0) == -0.1, self.c.get_last_err(is_abs=0)

    @unittest.skip
    def test_get_acc(self):
        """
        TODO: need this.
        :return:
        """
        pass

    def test_to_dict(self):
        pred_list = [1, 1, 1, 1, 1, 1, 1]
        true_list = [1, 1, 1, 1, 1, 1, 1]
        c = Counter(0)

        for pred, true in zip(pred_list, true_list):
            c.add_val(pred, true)

        mdict = c.metrics_to_dict()
        keys = ["epoch", "normal_acc", "strict_acc", "MSE", "MAE", "RMSE", "MAX"]
        for key in mdict.keys():
            assert key in keys


class CalTest(unittest.TestCase):
    def test_cal_acc(self):
        res = cal_acc([0.1, 0.1, 0.1, 0, 0], strict=1)
        assert res == 1.0, res

        res = cal_acc([0.1, 0.1, 0.1, 0, 0], strict=0)
        assert res == 1.0, res

        res = cal_acc([0.2, 0.1, 0.1, 0, 0], strict=0)
        assert res == 1.0

        res = cal_acc([0.2, 0.1, 0.1, 0, 0], strict=1)
        assert res == 0.8

        res = cal_acc([1.2, 0.1, 0.1, 0, 0], strict=0)
        assert res == 0.8

        res = cal_acc([1.2, 0.1, 0.1, 0, 0], strict=1)
        assert res == 0.8

        res = cal_acc([-1.2, 0.1, -0.1, 0, 0], strict=1)
        assert res == 0.8



class NumericAccuracy(object):
    def __init__(self, epoch):
        self.epoch = epoch
        self.pred_list = []
        self.true_list = []
        self.err_list = []
        self.cnt = 0

    def get_mae(self):
        err_list = [abs(val) for val in self.err_list]
        sums = sum(err_list)
        if not err_list:
            return 0
        assert self.cnt == len(self.pred_list)
        return sums / self.cnt

    def get_mse(self):
        sums = 0

        for err_val in self.err_list:
            sums += err_val**2

        if sums == 0:
            return 0
        return sums / self.cnt

    def get_rmse(self):
        return self.get_mse() ** 0.5

    def get_max_error(self):
        err_list = [val for val in self.err_list]
        return max(err_list)

    def get_max_abs_err(self):
        if not self.err_list:
            return -1
        err_list = [abs(val) for val in self.err_list]
        return max(err_list)



import json
import os
import pathlib

import numpy as np
import seaborn as sns
from pylab import exp, plot, show

# from pkg.recorder.db.relation.helper import save_epoch_data
from .helper import get_filename


class PredictResults:
    def __init__(self, parent_dir, prefix, run_name):
        self.cdf = list()
        self.ids = list()
        self.pred_list = list()
        self.true_list = list()

        self.run_name = run_name
        self.parent_dir = pathlib.Path(parent_dir)

        self.update_results_dir(prefix)
        self._create_dirs()

    def update_results_dir(self, prefix):
        self.cdf_dir = self.parent_dir / 'cdf' / prefix

    def _create_dirs(self):
        os.makedirs(self.parent_dir / "cdf", exist_ok=True)
        os.makedirs(self.cdf_dir, exist_ok=True)

    def insert(self, err_val, cur_id, pred, true, epoch):
        self.cdf.append(err_val)
        self.ids.append(str(cur_id))
        self.pred_list.append(str(pred))
        self.true_list.append(str(true))
        self._save2db(epoch, pred, true, err_val, cur_id)

    def _save_graph(self, epoch):
        """
        draw cdf graph.
        :param epoch:
        :return:
        """
        sns.set(color_codes=True)
        sns_plot = sns.distplot(
            tuple(self.cdf),
            bins=20,
            hist_kws=dict(cumulative=True),
            kde_kws=dict(cumulative=True),
        )
        fig = sns_plot.get_figure()
        fig.savefig(self.cdf_dir / (get_filename(epoch=epoch) + ".png"))

    def _save2db(self, epoch, pred, true, cdf, img_id):
        data = dict(
            run_name=self.run_name,
            epoch=epoch,
            pred=pred,
            true=true,
            cdf=cdf,
            img_id=img_id,
        )
        # save_epoch_data(data)

    def _save_data_json(self, epoch):
        """
        save to output dir, with json.
        :param epoch:
        :return:
        """
        filename = get_filename(epoch=epoch)
        data = dict(
            cdf=self.cdf, ids=self.ids, pred=self.pred_list, true=self.true_list
        )
        with open(self.cdf_dir / filename, "w") as outfile:
            json.dump(data, outfile)

    def save(self, epoch):
        self._save_data_json(epoch=epoch)
        self._save_graph(epoch=epoch)


def draw_cdf():
    dx = 0.01
    X = np.arange(-2, 2, dx)
    Y = exp(-X ** 2)

    # Normalize the data to a proper PDF
    Y /= (dx * Y).sum()

    # Compute the CDF
    CY = np.cumsum(Y * dx)

    plot(X, Y)
    plot(X, CY, "r--")

    show()
